package com.starbuds.server.service.statistics.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.convert.Convert;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.constant.statistics.StatTimeType;
import com.starbuds.server.common.pojo.db.StatChannelDO;
import com.starbuds.server.common.pojo.db.StatPlatformDO;
import com.starbuds.server.common.pojo.user.ChannelRatio;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.common.service.user.ChannelService;
import com.starbuds.server.service.statistics.provider.StatChannelProvider;
import com.starbuds.server.service.statistics.provider.StatPlatformProvider;
import com.starbuds.server.service.statistics.provider.StatSmsProvider;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.beans.BeanUtils;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.stream.Collectors;

/**
 * 平台数据统计服务实现
 *
 * @author Lay
 * @date 2020/2/5
 */
@Service(version = "1.0.0")
@Slf4j
public class StatPlatformServiceImpl implements StatPlatformService {

    @Resource
    StatPlatformProvider statPlatformProvider;

    @Reference(version = "1.0.0")
    ChannelService channelService;

    @Resource
    StatChannelProvider statChannelProvider;

    @Resource
    StatSmsProvider statSmsProvider;

    /**
     * 增加统计数
     *
     * @param channelId
     * @param statPlatformType
     * @param statValue
     */
    @Override
    public void addPlatformStat(Long channelId, StatPlatformTypeEnum statPlatformType, BigDecimal statValue) {
        if (statPlatformType == null) {
            log.warn("未知的统计类型");
            return;
        }

        addPlatformStat(channelId, statPlatformType.getValue(), statValue);
    }

    /**
     * 增加统计数
     *
     * @param channelId
     * @param statPlatformType
     * @param statValue
     */
    @Override
    public void addPlatformStat(Long channelId, int statPlatformType, BigDecimal statValue) {
        try {
            statPlatformProvider.addStatAsync(channelId, statPlatformType, statValue, System.currentTimeMillis());
        } catch (Exception e) {
            log.error("统计失败", e);
        }
    }

    @Override
    public boolean syncDataToChannel(Long startTime, Long endTime, Integer statTimeType) {

        log.info("开始同步数据到渠道，startTime:{},endTime:{},statTimeType:{}", startTime, endTime, statTimeType);

        try {
            // 统计类型
            List<Integer> statTypes = Arrays.stream(StatPlatformTypeEnum.values()).map(StatPlatformTypeEnum::getValue).collect(Collectors.toList());

            List<StatPlatformDO> list = statPlatformProvider.getList(statTypes, startTime, endTime, statTimeType);
            List<StatChannelDO> statList = new ArrayList<>();

            if (!CollectionUtil.isEmpty(list)) {
                Set<Long> channelIds = list.stream().map(StatPlatformDO::getChannelId).collect(Collectors.toSet());
                // 渠道分成比例
                Map<Long, ChannelRatio> ratioMap = channelService.getRation(new ArrayList<>(channelIds));

                SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd");
                long curTime = System.currentTimeMillis();
                String today = sdf.format(new Date());

                for (StatPlatformDO item : list) {
                    StatChannelDO channelDO = new StatChannelDO();
                    BeanUtils.copyProperties(item, channelDO);

                    // 分成比例
                    ChannelRatio channelRatio = ratioMap.get(item.getChannelId());
                    if (channelRatio == null) {
                        channelRatio = new ChannelRatio();
                        channelRatio.setTodayInstallRatio(1D);
                        channelRatio.setTomorrowInstallRatio(1D);
                        channelRatio.setTodaySplitRatio(1D);
                        channelRatio.setTomorrowSplitRatio(1D);
                        channelRatio.setChannelCreateTime(curTime);
                    }

                    channelDO.setStatTimeType(StatTimeType.YEAR_MONTH_DAY);
                    channelDO.setStatTime(Convert.toLong(Convert.toStr(item.getStatTime()).substring(0, Convert.toStr(item.getStatTime()).length() - 2)));

                    if (item.getStatType() == StatPlatformTypeEnum.INSTALL_COUNT.getValue()) {
                        if (today.equals(sdf.format(channelRatio.getChannelCreateTime()))) {
                            channelDO.setStatValue(item.getStatValue().multiply(new BigDecimal(channelRatio.getTodayInstallRatio())));
                        } else {
                            channelDO.setStatValue(item.getStatValue().multiply(new BigDecimal(channelRatio.getTomorrowInstallRatio())));
                        }
                    } else if (item.getStatType() == StatPlatformTypeEnum.REGISTER_USERS.getValue()) {
                        if (today.equals(sdf.format(channelRatio.getChannelCreateTime()))) {
                            channelDO.setStatValue(item.getStatValue().multiply(new BigDecimal(channelRatio.getTodayInstallRatio())));
                        } else {
                            channelDO.setStatValue(item.getStatValue().multiply(new BigDecimal(channelRatio.getTomorrowInstallRatio())));
                        }
                    } else if (item.getStatType() == StatPlatformTypeEnum.RECHARGE_AMOUNT.getValue()) {
                        if (today.equals(sdf.format(channelRatio.getChannelCreateTime()))) {
                            channelDO.setStatValue(item.getStatValue().multiply(new BigDecimal(channelRatio.getTodaySplitRatio())));
                        } else {
                            channelDO.setStatValue(item.getStatValue().multiply(new BigDecimal(channelRatio.getTomorrowSplitRatio())));
                        }
                    } else {
                        channelDO.setStatValue(item.getStatValue().multiply(new BigDecimal(channelRatio.getTomorrowSplitRatio())));
                    }
                    statList.add(channelDO);
                }
            }

            if (!CollectionUtil.isEmpty(statList)) {
                statChannelProvider.saveAll(statList);
                return true;
            }
            return true;
        } catch (Exception e) {
            log.info("执行出错", e);
            return false;
        }
    }

    /**
     * 增加短信发送统计
     *
     * @param ip
     * @param deviceId
     */
    @Override
    public void addSmsSendStat(String ip, String deviceId) {
        statSmsProvider.addStatAsync(ip, deviceId);
    }

    /**
     * 获取 ip 今日短信发送量
     *
     * @param ip
     * @return
     */
    @Override
    public int getSmsIpCountToday(String ip) {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd");
        int sendDate = Convert.toInt(sdf.format(new Date()));
        return statSmsProvider.getSmsIpCountDate(ip, sendDate);
    }

    /**
     * 获取设备今日短信发送量
     *
     * @param deviceId
     * @return
     */
    @Override
    public int getSmsDeviceCountToday(String deviceId) {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd");
        int sendDate = Convert.toInt(sdf.format(new Date()));
        return statSmsProvider.getSmsDeviceCountDate(deviceId, sendDate);
    }
}
