package com.starbuds.server.service.statistics.impl;

import cn.hutool.core.collection.CollectionUtil;
import com.alibaba.fastjson.JSON;
import com.starbuds.server.common.constant.finance.PayStatusEnum;
import com.starbuds.server.common.constant.statistics.RiskManageDateTypeEnum;
import com.starbuds.server.common.constant.statistics.RiskManageStatusEnum;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.constant.statistics.StatTimeType;
import com.starbuds.server.common.pojo.api.PageList;
import com.starbuds.server.common.pojo.db.RiskManageDO;
import com.starbuds.server.common.pojo.db.StatPlatformDO;
import com.starbuds.server.common.service.data.RedisService;
import com.starbuds.server.common.service.statistics.StatRiskService;
import com.starbuds.server.common.utils.DateUtil;
import com.starbuds.server.service.statistics.provider.*;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.scheduling.annotation.Async;

import javax.annotation.Resource;
import java.lang.reflect.Field;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * 风控数据统计
 *
 * @Author lujiao
 * @DateTime 2021/7/6 14:55
 * @Description
 */
@Service(version = "1.0.0")
@Slf4j
public class StatRiskServiceImpl implements StatRiskService {

    @Resource
    StatRiskManageProvider statRiskManageProvider;

    @Resource
    StatPlatformProvider statPlatformProvider;

    @Resource
    PayOrderProvider payOrderProvider;

    @Resource
    SkillOrderProvider skillOrderProvider;

    @Resource
    UserNobleHisProvider userNobleHisProvider;

    @Resource
    FamilyWalletWithdrawHisProvider familyWalletWithdrawHisProvider;

    /**
     * 不需要对比的字段
     */
    private static final String WHITE_FIELD = "riskId,dateType,userLotteryWin,luckyGiftWin,luckyBagGiftWin,userLotteryConsume,LuckyGiftConsume,luckyBagGiftConsume,status,startDate,endDate,riskFiled,extraData,withdrawFamily,withdrawFamilyUser";

    private static Map<String, String> fieldMap = new HashMap<>();

    static {
        fieldMap.put("registerUser", "注册用户数");
        fieldMap.put("activeUser", "活跃用户数");
        fieldMap.put("rechargeAmount", "充值金额");
        fieldMap.put("rechargeUsers", "充值用户数");
        fieldMap.put("firstRechargeAmount", "首充金额");
        fieldMap.put("payUser", "充值用户");
        fieldMap.put("firstRechargeUsers", "首充用户数");
        fieldMap.put("payNums", "支付成功订单数量");
        fieldMap.put("ARPU", "ARPU");
        fieldMap.put("ARPPU", "ARPPU");
        fieldMap.put("withdrawFamily", "家族提现");
        fieldMap.put("withdrawFamilyUser", "用户提现");
        fieldMap.put("nobleCost", "贵族消费");
        fieldMap.put("playWithCost", "陪玩消费");
        fieldMap.put("rewardRate", "返奖率");
    }


    private static final BigDecimal FLOOR_OVER_MULTI = new BigDecimal("1.35");

    private static final BigDecimal FLOOR_DOWN_MULTI = new BigDecimal("0.65");

    private static final BigDecimal ONE_HUNDRED = new BigDecimal(100);

    private final static Integer DAY_OF_MILLION = 60 * 60 * 24 * 1000;

    private final static Integer WEEK_OF_MILLION = 7 * 60 * 60 * 24 * 1000;

    private static Long START_TIME = 1606492800000L;

    private final static String DATE_SIMPLE_FORMAT = "yyyyMMdd";


    @Reference(version = "1.0.0")
    private RedisService redisService;

    /**
     * 生成日数据
     */
    @Override
    public List<String> generateDayRiskData(Date date) {
        String key = "generateDayRiskData";
        boolean lock = redisService.tryLock(key, "", 10, TimeUnit.SECONDS);
        if (lock) {
            // 检查之前是否存在数据了, 防止重复插入
            return doGenerateDayRiskData(date);
        } else {
            return new ArrayList<>();
        }
    }

    private List<String> doGenerateDayRiskData(Date date) {
        Integer intDate = DateUtil.getIntDate(date);
        // 这里创建当日数据
        RiskManageDO riskManageDO = loadDataFromStatPlatform(intDate);
        riskManageDO.setStartDate(intDate);
        riskManageDO.setEndDate(intDate);
        riskManageDO = generateOtherData(riskManageDO, date);
        List<String> compareRes = new ArrayList<>();
        riskManageDO = compareDiff(riskManageDO, date, compareRes);
        // 这里进行对比
        if (statRiskManageProvider.checkExist(intDate, RiskManageDateTypeEnum.DAY)) {
            statRiskManageProvider.update(riskManageDO, intDate, RiskManageDateTypeEnum.DAY);
        } else {
            statRiskManageProvider.insert(riskManageDO);
        }
        if (riskManageDO.getStatus() == RiskManageStatusEnum.TO_CHECK.getCode()) {
            return compareRes;
        } else {
            return new ArrayList<>();
        }
    }

    /**
     * 生成周数据
     *
     * @param date
     */
    @Override
    public void generateWeekRiskData(Date date) {
        long weekStart = DateUtil.getWeekStart(date);
        long weekEnd = weekStart + WEEK_OF_MILLION - 1;
        Integer beginDateInt = DateUtil.getIntDate(new Date(weekStart));
        Integer weekEndDateInt = DateUtil.getIntDate(new Date(weekEnd));
        generateRangRiskData(beginDateInt, weekEndDateInt, RiskManageDateTypeEnum.WEEK);
    }

    /**
     * 生成月数据
     *
     * @param date
     */
    @Override
    public void generateMonthRiskData(Date date) {
        long monthStart = DateUtil.getMonthStart(date);
        long monthEnd = DateUtil.getMonthEnd(date);
        Integer beginDateInt = DateUtil.getIntDate(new Date(monthStart));
        Integer monthEndDateInt = DateUtil.getIntDate(new Date(monthEnd));
        generateRangRiskData(beginDateInt, monthEndDateInt, RiskManageDateTypeEnum.MONTH);
    }

    @Override
    @Async
    public void generateHistoryRiskData() {
        generateHistoryDayRiskData();
        generateHistoryWeekRiskData();
        generateHistoryMonthRiskData();
    }

    /**
     * 更新数据状态
     *
     * @param riskId
     * @param statusEnum
     * @return
     */
    @Override
    public boolean updateRiskDataStatus(Long riskId, RiskManageStatusEnum statusEnum) {
        // 首先更新状态, 然后更新周 / 月的数据
        boolean updateRes = statRiskManageProvider.updateStatus(riskId, statusEnum) > 0;
        if (updateRes) {
            // 更新状态
            RiskManageDO riskManageDO = statRiskManageProvider.getRiskById(riskId);
            Date changeDate = cn.hutool.core.date.DateUtil.parse(riskManageDO.getStartDate().toString(), DATE_SIMPLE_FORMAT);
            Integer weekStart = DateUtil.getIntDate(new Date(DateUtil.getWeekStart(changeDate)));
            Integer weekEnd = DateUtil.getIntDate(new Date(DateUtil.getWeekEnd(changeDate)));
            Integer monthStart = DateUtil.getIntDate(new Date(DateUtil.getMonthStart(changeDate)));
            Integer monthEnd = DateUtil.getIntDate(new Date(DateUtil.getMonthEnd(changeDate)));
            generateRangRiskData(weekStart, weekEnd, RiskManageDateTypeEnum.WEEK);
            generateRangRiskData(monthStart, monthEnd, RiskManageDateTypeEnum.MONTH);
        }
        return statRiskManageProvider.updateStatus(riskId, statusEnum) > 0;
    }

    /**
     * 获取吉尼斯数据
     *
     * @return
     */
    @Override
    public RiskManageDO getGuinnessData(String fieldName, Integer dateType) {
        return statRiskManageProvider.getGuinnessDate(fieldName, dateType);
    }

    /**
     * 获取异常风控数据
     *
     * @return
     */
    @Override
    public PageList<RiskManageDO> getRiskData(Integer pageIndex, Integer pageSize) {
        return statRiskManageProvider.getRiskData(pageIndex, pageSize);
    }


    /**
     * 生成历史日数据
     */
    public void generateHistoryDayRiskData() {
        // 从2020-11-28号项目上线开始
        Long beginDate = START_TIME;
        Integer dateInt;
        Long endDate = DateUtil.getDayBegin(new Date());
        while (beginDate < endDate) {
            dateInt = DateUtil.getIntDate(new Date(beginDate));
            // 如果之前不存在的话, 生成数据
            if (!statRiskManageProvider.checkExist(dateInt, RiskManageDateTypeEnum.DAY)) {
                doGenerateDayRiskData(new Date(beginDate));
                beginDate = beginDate + DAY_OF_MILLION;
            }
        }
    }

    /**
     * 生成历史周数据
     */
    public void generateHistoryWeekRiskData() {
        Long beginDate = START_TIME;
        // 设定初始值
        Integer beginDateInt;
        Integer weekEndDateInt;
        beginDate = DateUtil.getWeekStart(new Date(beginDate));
        Long weekEndDate = beginDate + WEEK_OF_MILLION - 1;
        Long curDate = System.currentTimeMillis();
        while (beginDate < curDate) {
            beginDateInt = DateUtil.getIntDate(new Date(beginDate));
            weekEndDateInt = DateUtil.getIntDate(new Date(weekEndDate));
            generateRangRiskData(beginDateInt, weekEndDateInt, RiskManageDateTypeEnum.WEEK);
            beginDate = beginDate + WEEK_OF_MILLION;
            weekEndDate = weekEndDate + WEEK_OF_MILLION;
        }
    }

    /**
     * 生成区间数据
     *
     * @param beginDateInt
     * @param weekEndDateInt
     */
    private void generateRangRiskData(Integer beginDateInt, Integer weekEndDateInt, RiskManageDateTypeEnum riskManageDateTypeEnum) {
        RiskManageDO riskManageDo = statRiskManageProvider.getRiskSummary(beginDateInt, weekEndDateInt);
        // 这里大部分数值都是可以相加的, 但是有写数值不能相加, 所以需要重新统计,比如是通过计算获取的
        riskManageDo = loadCalcManageDO(riskManageDo);
        if (!statRiskManageProvider.checkExist(beginDateInt, riskManageDateTypeEnum)) {
            riskManageDo.setStatus(RiskManageStatusEnum.NORMAL.getCode());
            riskManageDo.setStartDate(beginDateInt);
            riskManageDo.setDateType(riskManageDateTypeEnum.getCode());
            riskManageDo.setEndDate(weekEndDateInt);
            statRiskManageProvider.insert(riskManageDo);
        } else {
            statRiskManageProvider.update(riskManageDo, beginDateInt, riskManageDateTypeEnum);
        }
    }

    /**
     * 生成历史月数据
     */
    public void generateHistoryMonthRiskData() {
        Long beginDate = START_TIME;
        // 设定初始值
        Integer beginDateInt;
        Integer monthEndDateInt;
        beginDate = DateUtil.getMonthStart(new Date(beginDate));
        Long monthEndDate = DateUtil.getMonthEnd(new Date(beginDate));
        Long curDate = System.currentTimeMillis();
        while (beginDate < curDate) {
            beginDateInt = DateUtil.getIntDate(new Date(beginDate));
            monthEndDateInt = DateUtil.getIntDate(new Date(monthEndDate));
            generateRangRiskData(beginDateInt, monthEndDateInt, RiskManageDateTypeEnum.MONTH);
            beginDate = monthEndDate + 1;
            monthEndDate = DateUtil.getMonthEnd(new Date(beginDate));
        }
    }

    /**
     * 获取该日期前七日的平均数值
     *
     * @param date
     * @return
     */
    @Override
    public RiskManageDO getAvgData(Date date) {
        Integer beginDate = DateUtil.getIntDate(DateUtil.getDateBefore(date, 8));
        Integer endDate = DateUtil.getIntDate(DateUtil.getDateBefore(date, 1));
        return statRiskManageProvider.getRiskAvgData(beginDate, endDate);
    }

    /**
     * 获取前七日数据并做对比
     *
     * @param riskManageDO
     * @return
     */
    private RiskManageDO compareDiff(RiskManageDO riskManageDO, Date date, List<String> compareRes) {
        riskManageDO.setStatus(RiskManageStatusEnum.NORMAL.getCode());
        Integer beginDate = DateUtil.getIntDate(DateUtil.getDateBefore(date, 8));
        Integer endDate = DateUtil.getIntDate(DateUtil.getDateBefore(date, 1));
        RiskManageDO sevenDayAvg = statRiskManageProvider.getRiskAvgData(beginDate, endDate);
        if (sevenDayAvg == null) {
            return riskManageDO;
        }
        // 通过反射的方式进行对比
        Field[] fields = RiskManageDO.class.getDeclaredFields();
        StringBuilder stringBuilder = new StringBuilder();
        for (Field field : fields) {
            field.setAccessible(true);
            String fieldName = field.getName();
            if (WHITE_FIELD.indexOf(fieldName) > -1) {
                continue;
            }
            try {
                Object fieldVal = field.get(riskManageDO);
                Object fieldValOrg = field.get(sevenDayAvg);
                if (fieldVal != null && fieldValOrg != null) {
                    BigDecimal filedValDecimal = new BigDecimal(fieldVal.toString());
                    BigDecimal filedValOrgDecimal = new BigDecimal(fieldValOrg.toString());
                    if (filedValOrgDecimal.compareTo(BigDecimal.ZERO) > 0) {
                        BigDecimal floorOver = filedValOrgDecimal.multiply(FLOOR_OVER_MULTI);
                        BigDecimal floorDown = filedValOrgDecimal.multiply(FLOOR_DOWN_MULTI);
                        if (filedValDecimal.compareTo(floorOver) > 0 ||
                                filedValDecimal.compareTo(floorDown) < 0) {
                            stringBuilder.append(fieldName);
                            stringBuilder.append(",");
                            String realName = fieldMap.get(fieldName);
                            compareRes.add(realName + " 原始值: " + filedValDecimal.toPlainString() + " || 均值 : " + filedValOrgDecimal.toPlainString());
                        }
                    }
                }
            } catch (Exception e) {
                log.error("对比风控数据错误" + date, e);
            }
        }
        riskManageDO.setExtraData(JSON.toJSONString(sevenDayAvg));
        String riskFields = stringBuilder.toString();
        riskManageDO.setRiskFiled(riskFields);
        if (StringUtils.isNotBlank(riskFields)) {
            riskManageDO.setStatus(RiskManageStatusEnum.TO_CHECK.getCode());
        }
        return riskManageDO;
    }

    /**
     * 生成其他需要填充的数据
     *
     * @param riskManageDO
     * @param date
     * @return
     */
    private RiskManageDO generateOtherData(RiskManageDO riskManageDO, Date date) {
        // 付费用户
        long dayBegin = DateUtil.getDayBegin(date);
        long dayEnd = DateUtil.getDayEnd(date);
        riskManageDO.setPayUser(payOrderProvider.getUserPayCount(dayBegin, dayEnd, PayStatusEnum.SUCCESSED));
        // 支付成功订单数
        riskManageDO.setPayNums(payOrderProvider.getPayCount(dayBegin, dayEnd, PayStatusEnum.SUCCESSED));
        // 贵族消费
        riskManageDO.setNobleCost(diamondToMoney(userNobleHisProvider.getNobleStat(dayBegin, dayEnd)));
        // 陪玩消费（陪玩 改 开黑）
        riskManageDO.setPlayWithCost(diamondToMoney(skillOrderProvider.getSkillOrderStat(dayBegin, dayEnd)));
        // 家族提现
        riskManageDO.setWithdrawFamily(diamondToMoney(familyWalletWithdrawHisProvider.getFamilyWithdrawStat(dayBegin, dayEnd)));

        riskManageDO = loadCalcManageDO(riskManageDO);
        return riskManageDO;
    }

    /**
     * 钻石转人民币
     *
     * @param diamondAmount
     * @return
     */
    private Integer diamondToMoney(Long diamondAmount) {
        if (diamondAmount != null) {
            return new BigDecimal(diamondAmount).divide(ONE_HUNDRED, 0, RoundingMode.HALF_UP).intValue();
        } else {
            return 0;
        }
    }

    /**
     * 抽取需要公共计算才能获取的数据
     * ARPU/ ARPPU/ rewardRate
     *
     * @param riskManageDO
     * @return
     */
    private RiskManageDO loadCalcManageDO(RiskManageDO riskManageDO) {
        // ARPU = 充值金额/ 活跃用户
        if (biggerThanZero(riskManageDO.getActiveUser()) && biggerThanZero(riskManageDO.getRechargeAmount())) {
            riskManageDO.setARPU((int) (riskManageDO.getRechargeAmount() / riskManageDO.getActiveUser()));
        }
        // ARPPU = 充值金额/ 付费用户
        if (biggerThanZero(riskManageDO.getPayUser()) && biggerThanZero(riskManageDO.getRechargeAmount())) {
            riskManageDO.setARPPU((int) (riskManageDO.getRechargeAmount() / riskManageDO.getPayUser()));
        }
        // 综合返奖率: (游戏中奖钻石 + 幸运礼物中奖钻石 + 背包幸运礼物中奖钻石) / (游戏消耗钻石 + (幸运礼物消耗钻石 + 背包幸运礼物消耗钻石) * 0.9)
        // 分子
        BigDecimal rateNumerator = new BigDecimal(riskManageDO.getUserLotteryWin() + riskManageDO.getLuckyBagGiftWin() + riskManageDO.getLuckyBagGiftWin());
        BigDecimal rateDenominator = new BigDecimal(riskManageDO.getUserLotteryConsume())
                .add(new BigDecimal(riskManageDO.getLuckyGiftConsume() + riskManageDO.getLuckyBagGiftConsume()).multiply(new BigDecimal("0.9")));
        if (rateDenominator.compareTo(BigDecimal.ZERO) > 0) {
            riskManageDO.setRewardRate(rateNumerator.multiply(ONE_HUNDRED).divide(rateDenominator, 2, RoundingMode.HALF_UP));
        } else {
            riskManageDO.setRewardRate(BigDecimal.ZERO);
        }
        return riskManageDO;
    }

    /**
     * 检查不为空且大于0
     *
     * @param input
     * @return
     */
    private boolean biggerThanZero(Integer input) {
        if (input != null && input > 0) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * 检查不为空且大于0
     *
     * @param input
     * @return
     */
    private boolean biggerThanZero(Long input) {
        if (input != null && input > 0) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * 数据填充
     *
     * @param intDate
     * @return
     */
    private RiskManageDO loadDataFromStatPlatform(Integer intDate) {
        RiskManageDO riskManageDO = new RiskManageDO();
        Long longDate = Long.valueOf(intDate);
        riskManageDO.setDateType(RiskManageDateTypeEnum.DAY.code());

        List<Integer> statTypes = Arrays.asList(new Integer[]{
                // 注册用户
                StatPlatformTypeEnum.REGISTER_USERS.getValue(),
                // 活跃用户
                StatPlatformTypeEnum.ACTIVE_USERS.getValue(),
                // 充值金额
                StatPlatformTypeEnum.RECHARGE_AMOUNT.getValue(),
                // 充值用户数
                StatPlatformTypeEnum.RECHARGE_USERS.getValue(),
                // 首充金额
                StatPlatformTypeEnum.FIRST_RECHARGE_AMOUNT.getValue(),
                // 首充用户数
                StatPlatformTypeEnum.FIRST_RECHARGE_USERS.getValue(),
                // 用户提现
                StatPlatformTypeEnum.WITHDRAW_USER.getValue(),
                // 抽奖活动中奖价值
                StatPlatformTypeEnum.LUCKY_ACTIVITY_WIN_AMOUNT.getValue(),
                // 幸运礼物中奖金额
                StatPlatformTypeEnum.LUCKY_GIFT_WIN_AMOUNT.getValue(),
                // 背包幸运礼物中奖金币
                StatPlatformTypeEnum.LUCKY_GIFT_WIN_AMOUNT_BAG.getValue(),
                // 抽奖活动消耗金币
                StatPlatformTypeEnum.LUCKY_ACTIVITY_PAY_AMOUNT.getValue(),
                // 幸运礼物消耗星钻
                StatPlatformTypeEnum.LUCKY_GIFT_PAY_AMOUNT.getValue(),
                // 背包幸运礼物消耗星钻
                StatPlatformTypeEnum.LUCKY_GIFT_PAY_AMOUNT_BAG.getValue()
        });
        List<StatPlatformDO> list = statPlatformProvider.selectSumByTypeAndTime(statTypes, longDate, StatTimeType.YEAR_MONTH_DAY);
        if (CollectionUtil.isNotEmpty(list)) {
            for (StatPlatformDO item : list) {
                if (item.getStatType() == StatPlatformTypeEnum.REGISTER_USERS.getValue()) {
                    riskManageDO.setRegisterUser(item.getStatValue().intValue());
                } else if (item.getStatType() == StatPlatformTypeEnum.ACTIVE_USERS.getValue()) {
                    riskManageDO.setActiveUser(item.getStatValue().intValue());
                } else if (item.getStatType() == StatPlatformTypeEnum.RECHARGE_AMOUNT.getValue()) {
                    riskManageDO.setRechargeAmount(item.getStatValue().longValue());
                } else if (item.getStatType() == StatPlatformTypeEnum.RECHARGE_USERS.getValue()) {
                    riskManageDO.setRechargeUsers(item.getStatValue().intValue());
                } else if (item.getStatType() == StatPlatformTypeEnum.FIRST_RECHARGE_AMOUNT.getValue()) {
                    riskManageDO.setFirstRechargeAmount(item.getStatValue().intValue());
                } else if (item.getStatType() == StatPlatformTypeEnum.FIRST_RECHARGE_USERS.getValue()) {
                    riskManageDO.setFirstRechargeUsers(item.getStatValue().intValue());
                } else if (item.getStatType() == StatPlatformTypeEnum.WITHDRAW_USER.getValue()) {
                    riskManageDO.setWithdrawFamilyUser(item.getStatValue().intValue());
                } else if (item.getStatType() == StatPlatformTypeEnum.LUCKY_ACTIVITY_WIN_AMOUNT.getValue()) {
                    riskManageDO.setUserLotteryWin(item.getStatValue().longValue());
                } else if (item.getStatType() == StatPlatformTypeEnum.LUCKY_GIFT_WIN_AMOUNT.getValue()) {
                    riskManageDO.setLuckyGiftWin(item.getStatValue().longValue());
                } else if (item.getStatType() == StatPlatformTypeEnum.LUCKY_GIFT_WIN_AMOUNT_BAG.getValue()) {
                    riskManageDO.setLuckyBagGiftWin(item.getStatValue().longValue());
                } else if (item.getStatType() == StatPlatformTypeEnum.LUCKY_ACTIVITY_PAY_AMOUNT.getValue()) {
                    riskManageDO.setUserLotteryConsume(item.getStatValue().longValue());
                } else if (item.getStatType() == StatPlatformTypeEnum.LUCKY_GIFT_PAY_AMOUNT.getValue()) {
                    riskManageDO.setLuckyGiftConsume(item.getStatValue().longValue());
                } else if (item.getStatType() == StatPlatformTypeEnum.LUCKY_GIFT_PAY_AMOUNT_BAG.getValue()) {
                    riskManageDO.setLuckyBagGiftConsume(item.getStatValue().longValue());
                }
            }
        }
        return riskManageDO;
    }


}
